@@ -7,8 +7,9 @@ from joblib import Parallel, delayed
 from sklearn.base import BaseEstimator, TransformerMixin
 from sklearn.utils.multiclass import unique_labels
 
-from .binning.functions import cat_processing, num_processing, refit_woe_dict
-from .model.functions import create_model, feature_select, generate_sql, predict_proba, save_reports
+from .binning.functions import cat_processing, find_cat_features, num_processing, prepare_data, refit
+from .model.functions import create_model, generate_sql, iv_feature_select, predict_proba, save_reports, \
+    sequential_feature_select
 
 
 class NpEncoder(json.JSONEncoder):
@@ -62,25 +63,15 @@ class WOETransformer(BaseEstimator, TransformerMixin):
         :param x: data matrix
         :param y: target vector
         """
-        if isinstance(x, pd.DataFrame):
-            if self.special_cols:
-                x = x.drop(self.special_cols, axis=1)
-            self.feature_names = x.columns
-        elif isinstance(x, np.ndarray):
-            self.feature_names = [f"X_{i}" for i in range(x.shape[-1])]
-        else:
-            raise TypeError("x vector is not np array neither data frame")
-
+        x, self.feature_names = prepare_data(x=x, special_cols=self.special_cols)
         self.classes_ = unique_labels(y)
 
         if len(self.cat_features) == 0 and self.cat_features_threshold > 0:
-            for i in range(len(self.feature_names)):
-                if (
-                        type(x[0, i]) == np.dtype("object")
-                        or type(x[0, i]) == np.dtype("str")
-                        or len(np.unique(x[:, i])) < self.cat_features_threshold
-                ):
-                    self.cat_features.append(self.feature_names[i])
+            self.cat_features = find_cat_features(
+                x=x,
+                feature_names=self.feature_names,
+                cat_features_threshold=self.cat_features_threshold
+            )
         if len(self.cat_features) > 0:
             self.num_features = [
                 feature
@@ -167,51 +158,28 @@ class WOETransformer(BaseEstimator, TransformerMixin):
             self.woe_iv_dict = json.load(file)
 
     def refit(self, x: pd.DataFrame, y: Union[pd.Series, np.ndarray]) -> None:
-        if isinstance(x, pd.DataFrame):
-            if self.special_cols:
-                x = x.drop(self.special_cols, axis=1)
-            self.feature_names = x.columns
-            self.feature_names = [col.replace("WOE_", "") for col in self.feature_names]
-        elif isinstance(x, np.ndarray):
-            self.feature_names = [f"X_{i}" for i in range(x.shape[-1])]
-        else:
-            raise TypeError("x vector is not np array neither data frame")
-
-        temp_woe_iv_dict = []
-
-        for i in range(len(self.woe_iv_dict)):
-            feature_idx = list(self.feature_names).index(
-                list(self.woe_iv_dict[i].keys())[0]
-            )
-            res_dict = refit_woe_dict(
-                x=x[:, feature_idx],
-                y=y,
-                bins=[
-                    _bin["bin"]
-                    for _bin in self.woe_iv_dict[i][list(self.woe_iv_dict[i].keys())[0]]
-                ],
-                type_feature=self.woe_iv_dict[i]["type_feature"],
-                missing_bin=self.woe_iv_dict[i]["missing_bin"]
-            )
-            temp_woe_iv_dict.append(
-                {
-                    list(self.woe_iv_dict[i].keys())[0]: res_dict,
-                    "missing_bin": self.woe_iv_dict[i]["missing_bin"],
-                    "type_feature": self.woe_iv_dict[i]["type_feature"],
-                }
-            )
-        self.woe_iv_dict = temp_woe_iv_dict
-        del temp_woe_iv_dict
+        x, self.feature_names = prepare_data(x=x, special_cols=self.special_cols)
+        self.woe_iv_dict = Parallel(n_jobs=self.n_jobs)(
+            delayed(refit)(
+                x[list(self.woe_iv_dict[i].keys())[0]],
+                y.values,
+                [_bin["bin"] for _bin in self.woe_iv_dict[i][list(self.woe_iv_dict[i].keys())[0]]],
+                self.woe_iv_dict[i]["type_feature"],
+                self.woe_iv_dict[i]["missing_bin"]
+            ) for i in range(len(self.woe_iv_dict))
+        )
 
 
 class CreateModel(BaseEstimator, TransformerMixin):
     def __init__(
             self,
-            max_vars: Union[int, float] = 0.8,
+            selection_method: str = 'iv',
+            max_vars: Union[int, float, None] = None,
             special_cols: List = None,
             unused_cols: List = None,
             n_jobs: int = None,
             gini_threshold: float = 5.0,
+            iv_threshold: float = 0.05,
             corr_threshold: float = 0.5,
             random_state: int = None,
             class_weight: str = None,
@@ -222,11 +190,13 @@ class CreateModel(BaseEstimator, TransformerMixin):
     ):
 
         self.results = None
+        self.selection_method = selection_method
         self.max_vars = max_vars
         self.special_cols = special_cols or []
         self.unused_cols = unused_cols or []
         self.n_jobs = n_jobs
         self.gini_threshold = gini_threshold
+        self.iv_threshold = iv_threshold
         self.corr_threshold = corr_threshold
         self.random_state = random_state
         self.class_weight = class_weight
@@ -241,31 +211,39 @@ class CreateModel(BaseEstimator, TransformerMixin):
         self.model = None
 
     def fit(self, x: pd.DataFrame, y: Union[pd.Series, np.ndarray]):
-        if isinstance(x, pd.DataFrame):
-            x = x.drop(self.special_cols + self.unused_cols, axis=1)
-            self.feature_names_ = x.columns
-        elif isinstance(x, np.ndarray):
-            self.feature_names_ = [f"X_{i}" for i in range(x.shape[-1])]
-        else:
-            raise TypeError("x vector is not np array neither data frame")
+        x, self.feature_names_ = prepare_data(x, special_cols=self.special_cols)
 
         if self.C is None:
             self.C = 1.0e4 / x.shape[0]
 
-        self.feature_names_ = feature_select(
-            x, y,
-            feature_names=self.feature_names_,
-            gini_threshold=self.gini_threshold,
-            corr_threshold=self.corr_threshold,
-            random_state=self.random_state,
-            class_weight=self.class_weight,
-            max_vars=self.max_vars,
-            direction=self.direction,
-            cv=self.cv,
-            c=self.C,
-            scoring=self.scoring,
-            n_jobs=self.n_jobs,
-        )
+        if self.max_vars is not None and self.max_vars < 1:
+            self.max_vars = int(len(self.feature_names_) * self.max_vars)
+
+        if self.selection_method == 'iv':
+            self.feature_names_ = iv_feature_select(
+                x, y,
+                feature_names=self.feature_names_,
+                iv_threshold=self.iv_threshold,
+                max_vars=self.max_vars,
+                n_jobs=self.n_jobs,
+            )
+        elif self.selection_method == 'sequential':
+            self.feature_names_ = sequential_feature_select(
+                x, y,
+                feature_names=self.feature_names_,
+                gini_threshold=self.gini_threshold,
+                corr_threshold=self.corr_threshold,
+                random_state=self.random_state,
+                class_weight=self.class_weight,
+                max_vars=self.max_vars,
+                direction=self.direction,
+                cv=self.cv,
+                c=self.C,
+                scoring=self.scoring,
+                n_jobs=self.n_jobs,
+            )
+        else:
+            raise NameError("selection_method should be 'iv' or 'sequential'")
 
         self.model = create_model(
             x, y,
